# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import time

from twisted.internet import error
from twisted.trial import unittest

from buildbot.process import remotetransfer
from buildbot.process.buildstep import create_step_from_step_or_factory
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.steps.source import cvs
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectCpdir
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.steps import ExpectUploadFile
from buildbot.test.util import sourcesteps


class TestCVS(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def setup_step(self, step, *args, **kwargs):
        super().setup_step(step, *args, **kwargs)

        # make parseGotRevision return something consistent, patching the class
        # instead of the object since a new object is constructed by runTest.
        def parseGotRevision(self):
            self.updateSourceProperty('got_revision', '2012-09-09 12:00:39 +0000')

        self.patch(cvs.CVS, 'parseGotRevision', parseGotRevision)

    def test_parseGotRevision(self):
        def gmtime():
            return time.struct_time((2012, 9, 9, 12, 9, 33, 6, 253, 0))

        self.patch(time, 'gmtime', gmtime)

        step = create_step_from_step_or_factory(
            cvs.CVS(cvsroot="x", cvsmodule="m", mode='full', method='clean')
        )
        props = []

        def updateSourceProperty(prop, name):
            props.append((prop, name))

        step.updateSourceProperty = updateSourceProperty

        step.parseGotRevision()
        self.assertEqual(props, [('got_revision', '2012-09-09 12:09:33 +0000')])

    def test_cvsEntriesContainStickyDates(self):
        step = cvs.CVS(cvsroot="x", cvsmodule="m", mode='full', method='clean')
        self.assertEqual(step._cvsEntriesContainStickyDates('D'), False)
        self.assertEqual(
            step._cvsEntriesContainStickyDates(
                '/file/1.1/Fri May 17 23:20:00//TMOZILLA_1_0_0_BRANCH\nD'
            ),
            False,
        )
        self.assertEqual(
            step._cvsEntriesContainStickyDates(
                '/file/1.1/Fri May 17 23:20:00//D2013.10.08.11.20.33\nD'
            ),
            True,
        )
        self.assertEqual(
            step._cvsEntriesContainStickyDates(
                '/file1/1.1/Fri May 17 23:20:00//\n'
                '/file2/1.1.2.3/Fri May 17 23:20:00//D2013.10.08.11.20.33\nD'
            ),
            True,
        )

    def test_mode_full_clean_and_login(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
                login="a password",
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    'login',
                ],
                initial_stdin="a password\n",
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS, state_string="update")
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_and_login_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
                login="a password",
            )
        )

        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    'login',
                ],
                initial_stdin="a password\n",
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS, state_string="update")
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_patch(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                workerdest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_patch_worker_2_16(self):
        self.setup_build(worker_version={'*': '2.16'})
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            ),
            patch=(1, 'patch'),
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                slavesrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-diff',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectDownloadFile(
                blocksize=32768,
                maxsize=None,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                slavedest='.buildbot-patched',
                workdir='wkdir',
                mode=None,
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'patch',
                    '-p1',
                    '--remove-empty-files',
                    '--force',
                    '--forward',
                    '-i',
                    '.buildbot-diff',
                ],
            ).exit(0),
            ExpectRmdir(dir='wkdir/.buildbot-diff', log_environ=True).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_timeout(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
                timeout=1,
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', timeout=1, command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', timeout=1, command=['cvsdiscard']).exit(0),
            ExpectShell(workdir='wkdir', timeout=1, command=['cvs', '-z3', 'update', '-dP']).exit(
                0
            ),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_branch(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
                branch='branch',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['cvs', '-z3', 'update', '-dP', '-r', 'branch']
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_branch_sourcestamp(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            ),
            args={'branch': 'my_branch'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard']).exit(0),
            ExpectShell(
                workdir='wkdir', command=['cvs', '-z3', 'update', '-dP', '-r', 'my_branch']
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_fresh(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='fresh',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard', '--ignore']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clobber(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='full',
            method='clobber',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clobber_retry(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='full',
            method='clobber',
            retry=(0, 2),
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_copy(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='full',
            method='copy',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='source/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='source/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='source/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='source', command=['cvs', '-z3', 'update', '-dP']).exit(0),
            ExpectCpdir(
                fromdir='source', todir='wkdir', log_environ=True, timeout=step.timeout
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_copy_wrong_repo(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='full',
            method='copy',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='source/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('the-end-of-the-universe\n')
            .exit(0),
            ExpectRmdir(dir='source', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'source',
                    'mozilla/browser/',
                ],
            ).exit(0),
            ExpectCpdir(
                fromdir='source', todir='wkdir', log_environ=True, timeout=step.timeout
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_sticky_date(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//D2013.10.08.11.20.33\nD\n')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_password_windows(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:dustin:secrets@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            # on Windows, this file does not contain the password, per
            # http://trac.buildbot.net/ticket/2355
            .upload_string(':pserver:dustin@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
                branch='my_branch',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['cvs', '-z3', 'update', '-dP', '-r', 'my_branch']
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_special_case(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
                branch='HEAD',
            ),
            args={"revision": '2012-08-16 16:05:16 +0000'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'cvs',
                    '-z3',
                    'update',
                    '-dP',
                    # note, no -r HEAD here - that's the special
                    # case
                    '-D',
                    '2012-08-16 16:05:16 +0000',
                ],
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mode_incremental_branch_sourcestamp(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
            ),
            args={'branch': 'my_branch'},
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(
                workdir='wkdir', command=['cvs', '-z3', 'update', '-dP', '-r', 'my_branch']
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_not_loggedin(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_no_existing_repo(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_retry(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
            retry=(0, 1),
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_wrong_repo(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('the-end-of-the-universe\n')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_wrong_module(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('the-end-of-the-universe\n')
            .exit(0),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_no_existing_repo(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            ).exit(1),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_clean_wrong_repo(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('the-end-of-the-universe\n')
            .exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_full_no_method(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard', '--ignore']).exit(0),
            ExpectShell(workdir='wkdir', command=['cvs', '-z3', 'update', '-dP']).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_with_options(self):
        step = cvs.CVS(
            cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
            cvsmodule="mozilla/browser/",
            mode='incremental',
            global_options=['-q'],
            extra_options=['-l'],
        )
        self.setup_step(step)
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            ).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=True, timeout=step.timeout).exit(0),
            ExpectShell(
                workdir='',
                command=[
                    'cvs',
                    '-q',
                    '-d',
                    ':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot',
                    '-z3',
                    'checkout',
                    '-d',
                    'wkdir',
                    '-l',
                    'mozilla/browser/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_mode_incremental_with_env_log_environ(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
                env={'abc': '123'},
                logEnviron=False,
            )
        )
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['cvs', '--version'], env={'abc': '123'}, log_environ=False
            ).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=False).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=['cvs', '-z3', 'update', '-dP'],
                env={'abc': '123'},
                log_environ=False,
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS)
        self.expect_property('got_revision', '2012-09-09 12:00:39 +0000', 'CVS')
        return self.run_step()

    def test_command_fails(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='incremental',
            )
        )
        self.expect_commands(ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(128))

        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_cvsdiscard_fails(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='fresh',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).exit(0),
            ExpectStat(file='wkdir/.buildbot-patched', log_environ=True).exit(1),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Root',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string(':pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Repository',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('mozilla/browser/\n')
            .exit(0),
            ExpectUploadFile(
                blocksize=32768,
                maxsize=None,
                workersrc='Entries',
                workdir='wkdir/CVS',
                writer=ExpectRemoteRef(remotetransfer.StringFileWriter),
            )
            .upload_string('/file/1.1/Fri May 17 23:20:00//\nD\n')
            .exit(0),
            ExpectShell(workdir='wkdir', command=['cvsdiscard', '--ignore'])
            .stderr('FAIL!\n')
            .exit(1),
        )

        self.expect_outcome(result=FAILURE)
        return self.run_step()

    def test_worker_connection_lost(self):
        self.setup_step(
            cvs.CVS(
                cvsroot=":pserver:anonymous@cvs-mirror.mozilla.org:/cvsroot",
                cvsmodule="mozilla/browser/",
                mode='full',
                method='clean',
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cvs', '--version']).error(error.ConnectionLost())
        )

        self.expect_outcome(result=RETRY, state_string="update (retry)")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Change(Row):
    table = "changes"

    id_column = 'changeid'

    def __init__(
        self,
        changeid=None,
        author='frank',
        committer='steve',
        comments='test change',
        branch='master',
        revision='abcd',
        revlink='http://vc/abcd',
        when_timestamp=1200000,
        category='cat',
        repository='repo',
        codebase='',
        project='proj',
        sourcestampid=92,
        parent_changeids=None,
    ):
        super().__init__(
            changeid=changeid,
            author=author,
            committer=committer,
            comments=comments,
            branch=branch,
            revision=revision,
            revlink=revlink,
            when_timestamp=when_timestamp,
            category=category,
            repository=repository,
            codebase=codebase,
            project=project,
            sourcestampid=sourcestampid,
            parent_changeids=parent_changeids,
        )


class ChangeFile(Row):
    table = "change_files"

    def __init__(self, changeid=None, filename=None):
        super().__init__(changeid=changeid, filename=filename)


class ChangeProperty(Row):
    table = "change_properties"

    def __init__(self, changeid=None, property_name=None, property_value=None):
        super().__init__(
            changeid=changeid, property_name=property_name, property_value=property_value
        )


class ChangeUser(Row):
    table = "change_users"

    def __init__(self, changeid=None, uid=None):
        super().__init__(changeid=changeid, uid=uid)



#! /usr/bin/python

# This is a script which delivers Change events from Darcs to the buildmaster
# each time a patch is pushed into a repository. Add it to the 'apply' hook
# on your canonical "central" repository, by putting something like the
# following in the _darcs/prefs/defaults file of that repository:
#
#  apply posthook /PATH/TO/darcs_buildbot.py BUILDMASTER:PORT
#  apply run-posthook
#
# (the second command is necessary to avoid the usual "do you really want to
# run this hook" prompt. Note that you cannot have multiple 'apply posthook'
# lines: if you need this, you must create a shell script to run all your
# desired commands, then point the posthook at that shell script.)
#
# Note that both Buildbot and Darcs must be installed on the repository
# machine. You will also need the Python/XML distribution installed (the
# "python2.3-xml" package under debian).


import os
import subprocess
import sys
import xml
from xml.dom import minidom

from twisted.internet import defer
from twisted.internet import reactor

from buildbot.clients import sendchange


def getText(node):
    return "".join([cn.data for cn in node.childNodes if cn.nodeType == cn.TEXT_NODE])


def getTextFromChild(parent, childtype):
    children = parent.getElementsByTagName(childtype)
    if not children:
        return ""
    return getText(children[0])


def makeChange(p):
    author = p.getAttribute("author")
    revision = p.getAttribute("hash")
    comments = getTextFromChild(p, "name") + "\n" + getTextFromChild(p, "comment")

    summary = p.getElementsByTagName("summary")[0]
    files = []
    for filenode in summary.childNodes:
        if filenode.nodeName in ("add_file", "modify_file", "remove_file"):
            filename = getText(filenode).strip()
            files.append(filename)
        elif filenode.nodeName == "move":
            to_name = filenode.getAttribute("to")
            files.append(to_name)

    # note that these are all unicode. Because PB can't handle unicode, we
    # encode them into ascii, which will blow up early if there's anything we
    # can't get to the far side. When we move to something that *can* handle
    # unicode (like newpb), remove this.
    author = author.encode("ascii", "replace")
    comments = comments.encode("ascii", "replace")
    files = [f.encode("ascii", "replace") for f in files]
    revision = revision.encode("ascii", "replace")

    change = {
        # note: this is more likely to be a full email address, which would
        # make the left-hand "Changes" column kind of wide. The buildmaster
        # should probably be improved to display an abbreviation of the
        # username.
        'username': author,
        'revision': revision,
        'comments': comments,
        'files': files,
    }
    return change


def getChangesFromCommand(cmd, count):
    out = subprocess.check_output(cmd, shell=True)
    out = out.decode(sys.stdout.encoding)
    try:
        doc = minidom.parseString(out)
    except xml.parsers.expat.ExpatError as e:
        print("failed to parse XML")
        print(str(e))
        print("purported XML is:")
        print("--BEGIN--")
        print(out)
        print("--END--")
        sys.exit(1)

    c = doc.getElementsByTagName("changelog")[0]
    changes = []
    for i, p in enumerate(c.getElementsByTagName("patch")):
        if i >= count:
            break
        changes.append(makeChange(p))
    return changes


def getSomeChanges(count):
    cmd = f"darcs changes --last={count} --xml-output --summary"
    return getChangesFromCommand(cmd, count)


LASTCHANGEFILE = ".darcs_buildbot-lastchange"


def findNewChanges():
    if os.path.exists(LASTCHANGEFILE):
        f = open(LASTCHANGEFILE)
        lastchange = f.read()
        f.close()
    else:
        return getSomeChanges(1)
    lookback = 10
    while True:
        changes = getSomeChanges(lookback)
        # getSomeChanges returns newest-first, so changes[0] is the newest.
        # we want to scan the newest first until we find the changes we sent
        # last time, then deliver everything newer than that (and send them
        # oldest-first).
        for i, c in enumerate(changes):
            if c['revision'] == lastchange:
                newchanges = changes[:i]
                newchanges.reverse()
                return newchanges
        if 2 * lookback > 100:
            raise RuntimeError(
                "unable to find our most recent change "
                f"({lastchange}) in the last {lookback} changes"
            )
        lookback = 2 * lookback


def sendChanges(master):
    changes = findNewChanges()
    s = sendchange.Sender(master)

    d = defer.Deferred()
    reactor.callLater(0, d.callback, None)

    if not changes:
        print("darcs_buildbot.py: weird, no changes to send")
        return
    elif len(changes) == 1:
        print("sending 1 change to buildmaster:")
    else:
        print(f"sending {len(changes)} changes to buildmaster:")

    # the Darcs Source class expects revision to be a context, not a
    # hash of a patch (which is what we have in c['revision']).  For
    # the moment, we send None for everything but the most recent, because getting
    # contexts is Hard.

    # get the context for the most recent change
    latestcontext = subprocess.check_output("darcs changes --context", shell=True)
    latestcontext = latestcontext.decode(sys.stdout.encoding)
    changes[-1]['context'] = latestcontext

    def _send(res, c):
        branch = None
        print(" {}".format(c['revision']))
        return s.send(
            branch, c.get('context'), c['comments'], c['files'], c['username'], vc='darcs'
        )

    for c in changes:
        d.addCallback(_send, c)

    def printSuccess(res):
        num_changes = len(changes)
        if num_changes > 1:
            print(f"{num_changes} changes sent successfully")
        elif num_changes == 1:
            print("change sent successfully")
        else:
            print("no changes to send")

    def printFailure(why):
        print("change(s) NOT sent, something went wrong: " + str(why))

    d.addCallbacks(printSuccess, printFailure)
    d.addBoth(lambda _: reactor.stop)
    reactor.run()

    if changes:
        lastchange = changes[-1]['revision']
        f = open(LASTCHANGEFILE, "w")
        f.write(lastchange)
        f.close()


if __name__ == '__main__':
    MASTER = sys.argv[1]
    sendChanges(MASTER)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import TypeVar

from twisted.internet import defer
from twisted.python import log
from twisted.web import resource
from twisted.web import server
from twisted.web.error import Error

from buildbot.util import unicode2bytes

if TYPE_CHECKING:
    from buildbot.master import BuildMaster

_CR_LF_RE = re.compile(rb"[\r\n]+.*")
_T = TypeVar('_T')


def protect_redirect_url(url: bytes) -> bytes:
    return _CR_LF_RE.sub(b"", url)


class Redirect(Error):
    def __init__(self, url: str | bytes) -> None:
        super().__init__(302, b"redirect")
        self.url = protect_redirect_url(unicode2bytes(url))


class Resource(resource.Resource):
    # if this is true for a class, then instances will have their
    # reconfigResource(new_config) methods called on reconfig.
    needsReconfig = False

    # as a convenience, subclasses have a ``master`` attribute, a
    # ``base_url`` attribute giving Buildbot's base URL,
    # and ``static_url`` attribute giving Buildbot's static files URL

    @property
    def base_url(self) -> str:
        return self.master.config.buildbotURL

    def __init__(self, master: BuildMaster) -> None:
        super().__init__()
        self.master = master
        if self.needsReconfig and master is not None:
            master.www.resourceNeedsReconfigs(self)

    def reconfigResource(self, new_config: object) -> None:
        raise NotImplementedError

    def asyncRenderHelper(
        self,
        request: server.Request,
        _callable: Callable[[server.Request], _T],
        writeError: Callable[..., None] | None = None,
    ) -> int:
        def writeErrorDefault(msg: bytes, errcode: int = 400) -> None:
            request.setResponseCode(errcode)
            request.setHeader(b'content-type', b'text/plain; charset=utf-8')
            request.write(msg)
            request.finish()

        if writeError is None:
            writeError = writeErrorDefault
        try:
            d: defer.Deferred[Any] = defer.maybeDeferred(_callable, request)
        except Exception as e:
            d = defer.fail(e)

        @d.addCallback
        def finish(s: bytes | None) -> None:
            try:
                if s is not None:
                    request.write(s)
                request.finish()
            except RuntimeError:  # pragma: no cover
                # this occurs when the client has already disconnected; ignore
                # it (see #2027)
                log.msg("http client disconnected before results were sent")

        @d.addErrback
        def failHttpRedirect(f: defer.Failure) -> None:
            f.trap(Redirect)
            request.redirect(f.value.url)
            request.finish()
            return None

        @d.addErrback
        def failHttpError(f: defer.Failure) -> None:
            f.trap(Error)
            e = f.value
            message = unicode2bytes(e.message)
            writeError(message, errcode=int(e.status))

        @d.addErrback
        def fail(f: defer.Failure) -> None:
            log.err(f, 'While rendering resource:')
            try:
                writeError(b'internal error - see logs', errcode=500)
            except Exception:
                try:
                    request.finish()
                except Exception:
                    pass

        return server.NOT_DONE_YET


class RedirectResource(Resource):
    def __init__(self, master: BuildMaster, basepath: str) -> None:
        super().__init__(master)
        self.basepath = basepath

    def render(self, request: server.Request) -> str:
        redir = self.base_url + self.basepath
        request.redirect(protect_redirect_url(unicode2bytes(redir)))
        return redir



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os

from twisted.trial import unittest

from buildbot.scripts import restart
from buildbot.scripts import start
from buildbot.scripts import stop
from buildbot.test.util import dirs
from buildbot.test.util import misc


def mkconfig(**kwargs):
    config = {"quiet": False, "basedir": os.path.abspath('basedir')}
    config.update(kwargs)
    return config


class TestStop(misc.StdoutAssertionsMixin, dirs.DirsMixin, unittest.TestCase):
    def setUp(self):
        self.setUpDirs('basedir')
        with open(os.path.join('basedir', 'buildbot.tac'), "w", encoding='utf-8') as f:
            f.write("Application('buildmaster')")
        self.setUpStdoutAssertions()

    # tests

    def test_restart_not_basedir(self):
        self.assertEqual(restart.restart(mkconfig(basedir='doesntexist')), 1)
        self.assertInStdout('invalid buildmaster directory')

    def test_restart_stop_fails(self):
        self.patch(stop, 'stop', lambda config, wait: 1)
        self.assertEqual(restart.restart(mkconfig()), 1)

    def test_restart_stop_succeeds_start_fails(self):
        self.patch(stop, 'stop', lambda config, wait: 0)
        self.patch(start, 'start', lambda config: 1)
        self.assertEqual(restart.restart(mkconfig()), 1)

    def test_restart_succeeds(self):
        self.patch(stop, 'stop', lambda config, wait: 0)
        self.patch(start, 'start', lambda config: 0)
        self.assertEqual(restart.restart(mkconfig()), 0)
        self.assertInStdout('now restarting')

    def test_restart_succeeds_quiet(self):
        self.patch(stop, 'stop', lambda config, wait: 0)
        self.patch(start, 'start', lambda config: 0)
        self.assertEqual(restart.restart(mkconfig(quiet=True)), 0)
        self.assertWasQuiet()

    def test_restart_clean(self):
        self.patch(stop, 'stop', lambda config, wait: 0)
        self.patch(start, 'start', lambda config: 0)
        self.assertEqual(restart.restart(mkconfig(quiet=True, clean=True)), 0)
        self.assertWasQuiet()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.python import deprecate
from twisted.python import versions

from buildbot.db import NULL
from buildbot.db import base
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    from twisted.internet import defer


@dataclass
class BuildDataModel:
    buildid: int
    name: str
    length: int
    source: str
    value: bytes | None

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'BuildDataConnectorComponent getBuildData, getBuildDataNoValue, and getAllBuildDataNoValues '
                'no longer return BuildData as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), BuildDataModel)
class BuildDataDict(BuildDataModel):
    pass


class BuildDataConnectorComponent(base.DBConnectorComponent):
    def _insert_race_hook(self, conn):
        # called so tests can simulate a race condition during insertion
        pass

    def setBuildData(
        self, buildid: int, name: str, value: bytes, source: str
    ) -> defer.Deferred[None]:
        def thd(conn) -> None:
            build_data_table = self.db.model.build_data

            retry = True
            while retry:
                try:
                    self.db.upsert(
                        conn,
                        build_data_table,
                        where_values=(
                            (build_data_table.c.buildid, buildid),
                            (build_data_table.c.name, name),
                        ),
                        update_values=(
                            (build_data_table.c.value, value),
                            (build_data_table.c.length, len(value)),
                            (build_data_table.c.source, source),
                        ),
                        _race_hook=self._insert_race_hook,
                    )
                    conn.commit()
                except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                    # there's been a competing insert, retry
                    conn.rollback()
                    if not retry:
                        raise
                finally:
                    retry = False

        return self.db.pool.do(thd)

    def getBuildData(self, buildid: int, name: str) -> defer.Deferred[BuildDataModel | None]:
        def thd(conn) -> BuildDataModel | None:
            build_data_table = self.db.model.build_data

            q = build_data_table.select().where(
                (build_data_table.c.buildid == buildid) & (build_data_table.c.name == name)
            )
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            return self._model_from_row(row, value=row.value)

        return self.db.pool.do(thd)

    def getBuildDataNoValue(self, buildid: int, name: str) -> defer.Deferred[BuildDataModel | None]:
        def thd(conn) -> BuildDataModel | None:
            build_data_table = self.db.model.build_data

            q = sa.select(
                build_data_table.c.buildid,
                build_data_table.c.name,
                build_data_table.c.length,
                build_data_table.c.source,
            )
            q = q.where((build_data_table.c.buildid == buildid) & (build_data_table.c.name == name))
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            return self._model_from_row(row, value=None)

        return self.db.pool.do(thd)

    def getAllBuildDataNoValues(self, buildid: int) -> defer.Deferred[list[BuildDataModel]]:
        def thd(conn) -> list[BuildDataModel]:
            build_data_table = self.db.model.build_data

            q = sa.select(
                build_data_table.c.buildid,
                build_data_table.c.name,
                build_data_table.c.length,
                build_data_table.c.source,
            )
            q = q.where(build_data_table.c.buildid == buildid)

            return [self._model_from_row(row, value=None) for row in conn.execute(q).fetchall()]

        return self.db.pool.do(thd)

    def deleteOldBuildData(self, older_than_timestamp: int) -> defer.Deferred[int]:
        build_data = self.db.model.build_data
        builds = self.db.model.builds

        def count_build_datum(conn) -> int:
            res = conn.execute(sa.select(sa.func.count(build_data.c.id)))
            count = res.fetchone()[0]
            res.close()
            return count

        def thd(conn) -> int:
            count_before = count_build_datum(conn)

            if self.db._engine.dialect.name == 'sqlite':
                # sqlite does not support delete with a join, so for this case we use a subquery,
                # which is much slower

                q = sa.select(builds.c.id)
                q = q.where(
                    (builds.c.complete_at >= older_than_timestamp) | (builds.c.complete_at == NULL)
                )
                # n.b.: in sqlite we need to filter on `>= older_than_timestamp` because of the following `NOT IN` clause...

                q = build_data.delete().where(build_data.c.buildid.notin_(q))
            else:
                q = build_data.delete()
                q = q.where(builds.c.id == build_data.c.buildid)
                q = q.where(builds.c.complete_at < older_than_timestamp)
            res = conn.execute(q)
            conn.commit()
            res.close()

            count_after = count_build_datum(conn)
            return count_before - count_after

        return self.db.pool.do(thd)

    def _model_from_row(self, row, value: bytes | None):
        return BuildDataModel(
            buildid=row.buildid,
            name=row.name,
            length=row.length,
            source=row.source,
            value=value,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Buildbot plugin infrastructure
"""

from buildbot import statistics
from buildbot.interfaces import IBuildStep
from buildbot.interfaces import IChangeSource
from buildbot.interfaces import IScheduler
from buildbot.interfaces import IWorker
from buildbot.plugins.db import get_plugins

__all__ = [
    'changes',
    'reporters',
    'schedulers',
    'secrets',
    'statistics',
    'steps',
    'util',
    'webhooks',
    'worker',
]


# Names here match the names of the corresponding Buildbot module, hence
# 'changes', 'schedulers', but 'buildslave'
changes = get_plugins('changes', IChangeSource)
schedulers = get_plugins('schedulers', IScheduler)
steps = get_plugins('steps', IBuildStep)
util = get_plugins('util', None)
reporters = get_plugins('reporters', None)
secrets = get_plugins('secrets', None)
webhooks = get_plugins('webhooks', None)

# Worker entry point for new/updated plugins.
worker = get_plugins('worker', IWorker)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import pprint
import re
import textwrap

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.changes import changes
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class Change(TestReactorMixin, unittest.TestCase):
    change23_rows = [
        fakedb.SourceStamp(id=92),
        fakedb.Change(
            changeid=23,
            author="dustin",
            committer="dustin",
            comments="fix whitespace",
            branch="warnerdb",
            revision="deadbeef",
            when_timestamp=266738404,
            revlink='http://warner/0e92a098b',
            category='devel',
            repository='git://warner',
            codebase='mainapp',
            project='Buildbot',
        ),
        fakedb.ChangeFile(changeid=23, filename='master/README.txt'),
        fakedb.ChangeFile(changeid=23, filename='worker/README.txt'),
        fakedb.ChangeProperty(
            changeid=23, property_name='notest', property_value='["no","Change"]'
        ),
        fakedb.User(uid=27),
        fakedb.ChangeUser(changeid=23, uid=27),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.change23 = changes.Change(**{  # using **dict(..) forces kwargs
            "category": 'devel',
            "repository": 'git://warner',
            "codebase": 'mainapp',
            "who": 'dustin',
            "committer": 'dustin',
            "when": 266738404,
            "comments": 'fix whitespace',
            "project": 'Buildbot',
            "branch": 'warnerdb',
            "revlink": 'http://warner/0e92a098b',
            "properties": {'notest': "no"},
            "files": ['master/README.txt', 'worker/README.txt'],
            "revision": 'deadbeef',
        })
        self.change23.number = 23

        self.change24 = changes.Change(**{
            "category": 'devel',
            "repository": 'git://warner',
            "codebase": 'mainapp',
            "who": 'dustin',
            "committer": 'dustin',
            "when": 266738405,
            "comments": 'fix whitespace again',
            "project": 'Buildbot',
            "branch": 'warnerdb',
            "revlink": 'http://warner/0e92a098c',
            "properties": {'notest': "no"},
            "files": ['master/README.txt', 'worker/README.txt'],
            "revision": 'deadbeef',
        })
        self.change24.number = 24

        self.change25 = changes.Change(**{
            "category": 'devel',
            "repository": 'git://warner',
            "codebase": 'mainapp',
            "who": 'dustin',
            "committer": 'dustin',
            "when": 266738406,
            "comments": 'fix whitespace again',
            "project": 'Buildbot',
            "branch": 'warnerdb',
            "revlink": 'http://warner/0e92a098d',
            "properties": {'notest': "no"},
            "files": ['master/README.txt', 'worker/README.txt'],
            "revision": 'deadbeef',
        })
        self.change25.number = 25

    @defer.inlineCallbacks
    def test_fromChdict(self):
        # get a real honest-to-goodness chdict from the fake db
        yield self.master.db.insert_test_data(self.change23_rows)
        chdict = yield self.master.db.changes.getChange(23)

        exp = self.change23
        got = yield changes.Change.fromChdict(self.master, chdict)

        # compare
        ok = True
        ok = ok and got.number == exp.number
        ok = ok and got.who == exp.who
        ok = ok and got.committer == exp.committer
        ok = ok and sorted(got.files) == sorted(exp.files)
        ok = ok and got.comments == exp.comments
        ok = ok and got.revision == exp.revision
        ok = ok and got.when == exp.when
        ok = ok and got.branch == exp.branch
        ok = ok and got.category == exp.category
        ok = ok and got.revlink == exp.revlink
        ok = ok and got.properties == exp.properties
        ok = ok and got.repository == exp.repository
        ok = ok and got.codebase == exp.codebase
        ok = ok and got.project == exp.project
        if not ok:

            def printable(c):
                return pprint.pformat(c.__dict__)

            self.fail(f"changes do not match; expected\n{printable(exp)}\ngot\n{printable(got)}")

    def test_str(self):
        string = str(self.change23)
        self.assertTrue(re.match(r"Change\(.*\)", string), string)

    def test_asText(self):
        text = self.change23.asText()
        self.assertTrue(
            re.match(
                textwrap.dedent("""\
            Files:
             master/README.txt
             worker/README.txt
            On: git://warner
            For: Buildbot
            At: .*
            Changed By: dustin
            Committed By: dustin
            Comments: fix whitespaceProperties:.
              notest: no

            """),
                text,
            ),
            text,
        )

    def test_asDict(self):
        dict = self.change23.asDict()
        self.assertIn('1978', dict['at'])  # timezone-sensitive
        del dict['at']
        self.assertEqual(
            dict,
            {
                'branch': 'warnerdb',
                'category': 'devel',
                'codebase': 'mainapp',
                'comments': 'fix whitespace',
                'files': [{'name': 'master/README.txt'}, {'name': 'worker/README.txt'}],
                'number': 23,
                'project': 'Buildbot',
                'properties': [('notest', 'no', 'Change')],
                'repository': 'git://warner',
                'rev': 'deadbeef',
                'revision': 'deadbeef',
                'revlink': 'http://warner/0e92a098b',
                'when': 266738404,
                'who': 'dustin',
                'committer': 'dustin',
            },
        )

    def test_getShortAuthor(self):
        self.assertEqual(self.change23.getShortAuthor(), 'dustin')

    def test_getTime(self):
        # careful, or timezones will hurt here
        self.assertIn('Jun 1978', self.change23.getTime())

    def test_getTimes(self):
        self.assertEqual(self.change23.getTimes(), (266738404, None))

    def test_getText(self):
        self.change23.who = 'nasty < nasty'  # test the html escaping (ugh!)
        self.assertEqual(self.change23.getText(), ['nasty &lt; nasty'])

    def test_getLogs(self):
        self.assertEqual(self.change23.getLogs(), {})

    def test_compare(self):
        self.assertEqual(self.change23, self.change23)
        self.assertNotEqual(self.change24, self.change23)
        self.assertGreater(self.change24, self.change23)
        self.assertGreaterEqual(self.change24, self.change23)
        self.assertGreaterEqual(self.change24, self.change24)
        self.assertLessEqual(self.change24, self.change24)
        self.assertLessEqual(self.change23, self.change24)
        self.assertLess(self.change23, self.change25)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime
from unittest.mock import Mock

from dateutil.tz import tzutc
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.plugins import util
from buildbot.process.builder import Builder
from buildbot.process.properties import Interpolate
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.reporters.bitbucketserver import HTTP_CREATED
from buildbot.reporters.bitbucketserver import HTTP_PROCESSED
from buildbot.reporters.bitbucketserver import BitbucketServerCoreAPIStatusPush
from buildbot.reporters.bitbucketserver import BitbucketServerPRCommentPush
from buildbot.reporters.bitbucketserver import BitbucketServerStatusPush
from buildbot.reporters.generators.build import BuildStartEndStatusGenerator
from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.reporters.generators.buildset import BuildSetStatusGenerator
from buildbot.reporters.message import MessageFormatter
from buildbot.reporters.message import MessageFormatterRenderable
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.logging import LoggingMixin
from buildbot.test.util.reporter import ReporterTestMixin

HTTP_NOT_FOUND = 404


class TestException(Exception):
    pass


class TestBitbucketServerStatusPush(
    TestReactorMixin, ConfigErrorsMixin, unittest.TestCase, ReporterTestMixin, LoggingMixin
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)
        yield self.master.startService()

    @defer.inlineCallbacks
    def setupReporter(self, **kwargs):
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, 'serv', auth=('username', 'passwd'), debug=None, verify=None
        )
        self.sp = BitbucketServerStatusPush(
            "serv", Interpolate("username"), Interpolate("passwd"), **kwargs
        )
        yield self.sp.setServiceParent(self.master)
        self.addCleanup(self.master.stopService)

    @defer.inlineCallbacks
    def _check_start_and_finish_build(self, build):
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'INPROGRESS',
                'key': 'Builder0',
                'description': 'Build started.',
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'SUCCESSFUL',
                'key': 'Builder0',
                'description': 'Build done.',
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'FAILED',
                'key': 'Builder0',
                'description': 'Build done.',
            },
        )
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_basic(self):
        self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS)
        yield self._check_start_and_finish_build(build)

    @defer.inlineCallbacks
    def test_setting_options(self):
        generator = BuildStartEndStatusGenerator(
            start_formatter=MessageFormatterRenderable('Build started.'),
            end_formatter=MessageFormatterRenderable('Build finished.'),
        )

        self.setupReporter(statusName='Build', generators=[generator])
        build = yield self.insert_build_finished(SUCCESS)
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'INPROGRESS',
                'key': 'Builder0',
                'name': 'Build',
                'description': 'Build started.',
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'SUCCESSFUL',
                'key': 'Builder0',
                'name': 'Build',
                'description': 'Build finished.',
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'FAILED',
                'key': 'Builder0',
                'name': 'Build',
                'description': 'Build finished.',
            },
            code=HTTP_PROCESSED,
        )
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_error(self):
        self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS)
        self._http.expect(
            'post',
            '/rest/build-status/1.0/commits/d34db33fd43db33f',
            json={
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'state': 'INPROGRESS',
                'key': 'Builder0',
                'description': 'Build started.',
            },
            code=HTTP_NOT_FOUND,
            content_json={
                "error_description": "This commit is unknown to us",
                "error": "invalid_commit",
            },
        )
        build['complete'] = False
        self.setUpLogging()
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged('404: Unable to send Bitbucket Server status')

    @defer.inlineCallbacks
    def test_basic_with_no_revision(self):
        yield self.setupReporter()
        self.reporter_test_revision = None

        build = yield self.insert_build_finished(SUCCESS)

        self.setUpLogging()
        # we don't expect any request
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged("Unable to get the commit hash")
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)


class TestBitbucketServerCoreAPIStatusPush(
    ConfigErrorsMixin, TestReactorMixin, unittest.TestCase, ReporterTestMixin, LoggingMixin
):
    @defer.inlineCallbacks
    def setupReporter(self, token=None, **kwargs):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

        def setup_properties(props):
            props.setProperty("buildername", "Builder0", "Builder")
            return defer.succeed(None)

        builder = Mock(spec=Builder)
        builder.master = self.master
        builder.name = "Builder0"
        builder.setup_properties = setup_properties
        self.master.botmaster.getBuilderById = Mock(return_value=builder)

        http_headers = {} if token is None else {'Authorization': 'Bearer tokentoken'}
        http_auth = ('username', 'passwd') if token is None else None

        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, 'serv', auth=http_auth, headers=http_headers, debug=None, verify=None
        )

        auth = (Interpolate("username"), Interpolate("passwd")) if token is None else None

        self.sp = BitbucketServerCoreAPIStatusPush("serv", token=token, auth=auth, **kwargs)
        yield self.sp.setServiceParent(self.master)
        yield self.master.startService()

        @defer.inlineCallbacks
        def cleanup():
            if self.master.running:
                yield self.master.stopService()

        self.addCleanup(cleanup)

    def setUp(self):
        self.master = None

    @defer.inlineCallbacks
    def _check_start_and_finish_build(self, build, parentPlan=False, epoch=False):
        _name = "Builder_parent #1 \u00bb Builder0 #0" if parentPlan else "Builder0 #0"
        _parent = "Builder_parent" if parentPlan else "Builder0"

        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': _name,
                'description': 'Build started.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': 'refs/heads/master',
                'buildNumber': '0',
                'state': 'INPROGRESS',
                'parent': _parent,
                'duration': None,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': _name,
                'description': 'Build done.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': 'refs/heads/master',
                'buildNumber': '0',
                'state': 'SUCCESSFUL',
                'parent': _parent,
                'duration': 10000,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': _name,
                'description': 'Build done.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': 'refs/heads/master',
                'buildNumber': '0',
                'state': 'FAILED',
                'parent': _parent,
                'duration': 10000,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        if epoch:
            build['started_at'] = 1554161913
        else:
            build['started_at'] = datetime.datetime(2019, 4, 1, 23, 38, 33, 154354, tzinfo=tzutc())
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        if epoch:
            build["complete_at"] = 1554161923
        else:
            build["complete_at"] = datetime.datetime(2019, 4, 1, 23, 38, 43, 154354, tzinfo=tzutc())
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)
        build['results'] = FAILURE
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_buildrequest(self):
        yield self.setupReporter()
        buildrequest = yield self.insert_buildrequest_new()

        _name = "Builder0 #(build request)"
        _parent = "Builder0"
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': _name,
                'description': 'Build pending.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/buildrequests/11',
                'ref': 'refs/heads/master',
                'buildNumber': '',
                'state': 'INPROGRESS',
                'parent': _parent,
                'duration': None,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': _name,
                'description': 'Build pending.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/buildrequests/11',
                'ref': 'refs/heads/master',
                'buildNumber': '',
                'state': 'FAILED',
                'parent': _parent,
                'duration': None,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )

        yield self.sp._got_event(('buildrequests', 11, 'new'), buildrequest)
        yield self.sp._got_event(('buildrequests', 11, 'cancel'), buildrequest)

    def test_config_no_base_url(self):
        with self.assertRaisesConfigError("Parameter base_url has to be given"):
            BitbucketServerCoreAPIStatusPush(base_url=None)

    def test_config_auth_and_token_mutually_exclusive(self):
        with self.assertRaisesConfigError(
            "Only one authentication method can be given (token or auth)"
        ):
            BitbucketServerCoreAPIStatusPush("serv", token="x", auth=("username", "passwd"))

    @defer.inlineCallbacks
    def test_basic(self):
        yield self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS)
        yield self._check_start_and_finish_build(build)

    @defer.inlineCallbacks
    def test_basic_epoch(self):
        yield self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS)
        yield self._check_start_and_finish_build(build, epoch=True)

    @defer.inlineCallbacks
    def test_with_parent(self):
        yield self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS, parent_plan=True)
        yield self._check_start_and_finish_build(build, parentPlan=True)

    @defer.inlineCallbacks
    def test_with_token(self):
        yield self.setupReporter(token='tokentoken')
        build = yield self.insert_build_finished(SUCCESS)
        yield self._check_start_and_finish_build(build)

    @defer.inlineCallbacks
    def test_error_setup_status(self):
        yield self.setupReporter()

        @defer.inlineCallbacks
        def raise_deferred_exception(**kwargs):
            raise TestException()

        self.sp.createStatus = Mock(side_effect=raise_deferred_exception)
        build = yield self.insert_build_finished(SUCCESS)
        yield self.sp._got_event(('builds', 20, 'new'), build)

        self.assertEqual(len(self.flushLoggedErrors(TestException)), 1)

    @defer.inlineCallbacks
    def test_error(self):
        self.setupReporter()
        build = yield self.insert_build_finished(SUCCESS)
        self.setUpLogging()
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': 'Builder0 #0',
                'description': 'Build started.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': 'refs/heads/master',
                'buildNumber': '0',
                'state': 'INPROGRESS',
                'parent': 'Builder0',
                'duration': None,
                'testResults': None,
            },
            code=HTTP_NOT_FOUND,
        )
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged('404: Unable to send Bitbucket Server status')

    @defer.inlineCallbacks
    def test_with_full_ref(self):
        yield self.setupReporter()
        self.reporter_test_branch = "refs/heads/master"
        build = yield self.insert_build_finished(SUCCESS)

        yield self._check_start_and_finish_build(build)

    @defer.inlineCallbacks
    def test_with_no_ref(self):
        yield self.setupReporter()

        self.reporter_test_branch = None
        build = yield self.insert_build_finished(SUCCESS)

        self.setUpLogging()
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': 'Builder0 #0',
                'description': 'Build started.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': None,
                'buildNumber': '0',
                'state': 'INPROGRESS',
                'parent': 'Builder0',
                'duration': None,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged("WARNING: Unable to resolve ref for SSID: 234.")

    @defer.inlineCallbacks
    def test_with_no_revision(self):
        yield self.setupReporter()

        self.reporter_test_revision = None
        build = yield self.insert_build_finished(SUCCESS)

        self.setUpLogging()
        # we don't expect any request
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged("Unable to get the commit hash for SSID: 234")

    @defer.inlineCallbacks
    def test_with_no_repo(self):
        yield self.setupReporter()

        self.reporter_test_repo = ''
        build = yield self.insert_build_finished(SUCCESS)

        self.setUpLogging()
        # we don't expect any request
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged("Unable to parse repository info from '' for SSID: 234")

    @defer.inlineCallbacks
    def test_with_renderers(self):
        @util.renderer
        def r_testresults(props):
            return {
                "failed": props.getProperty("unittests_failed", 0),
                "skipped": props.getProperty("unittests_skipped", 0),
                "successful": props.getProperty("unittests_successful", 0),
            }

        @util.renderer
        def r_duration(props):
            return props.getProperty("unittests_runtime")

        yield self.setupReporter(
            statusName=Interpolate("%(prop:plan_name)s"),
            statusSuffix=Interpolate(" [%(prop:unittests_os)s]"),
            buildNumber=Interpolate('100'),
            ref=Interpolate("%(prop:branch)s"),
            parentName=Interpolate("%(prop:master_plan)s"),
            testResults=r_testresults,
            duration=r_duration,
        )

        self.reporter_test_props['unittests_failed'] = 0
        self.reporter_test_props['unittests_skipped'] = 2
        self.reporter_test_props['unittests_successful'] = 3
        self.reporter_test_props['unittests_runtime'] = 50000
        self.reporter_test_props['unittests_os'] = "win10"
        self.reporter_test_props['plan_name'] = "Unittests"
        self.reporter_test_props['master_plan'] = "Unittests-master"
        build = yield self.insert_build_finished(SUCCESS)

        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': 'Unittests [win10]',
                'description': 'Build done.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': "refs/pull/34/merge",
                'buildNumber': '100',
                'state': 'SUCCESSFUL',
                'parent': 'Unittests-master',
                'duration': 50000,
                'testResults': {'failed': 0, 'skipped': 2, 'successful': 3},
            },
            code=HTTP_PROCESSED,
        )
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_with_test_results(self):
        yield self.setupReporter()

        self.reporter_test_props['tests_skipped'] = 2
        self.reporter_test_props['tests_successful'] = 3
        build = yield self.insert_build_finished(SUCCESS)

        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': 'Builder0 #0',
                'description': 'Build done.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': 'refs/heads/master',
                'buildNumber': '0',
                'state': 'SUCCESSFUL',
                'parent': 'Builder0',
                'duration': 10000,
                'testResults': {'failed': 0, 'skipped': 2, 'successful': 3},
            },
            code=HTTP_PROCESSED,
        )
        build['started_at'] = datetime.datetime(2019, 4, 1, 23, 38, 33, 154354, tzinfo=tzutc())
        build["complete_at"] = datetime.datetime(2019, 4, 1, 23, 38, 43, 154354, tzinfo=tzutc())
        build['complete'] = True
        yield self.sp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_verbose(self):
        yield self.setupReporter(verbose=True)
        build = yield self.insert_build_finished(SUCCESS)
        self.setUpLogging()
        self._http.expect(
            'post',
            '/rest/api/1.0/projects/example.org/repos/repo/commits/d34db33fd43db33f/builds',
            json={
                'name': 'Builder0 #0',
                'description': 'Build started.',
                'key': 'Builder0',
                'url': 'http://localhost:8080/#/builders/79/builds/0',
                'ref': "refs/heads/master",
                'buildNumber': '0',
                'state': 'INPROGRESS',
                'parent': 'Builder0',
                'duration': None,
                'testResults': None,
            },
            code=HTTP_PROCESSED,
        )
        build['complete'] = False
        yield self.sp._got_event(('builds', 20, 'new'), build)
        self.assertLogged('Sending payload:')
        self.assertLogged('Status "INPROGRESS" sent for example.org/repo d34db33fd43db33f')


UNICODE_BODY = "body: \u00e5\u00e4\u00f6 text"
EXPECTED_API = '/rest/api/1.0/projects/PRO/repos/myrepo/pull-requests/20/comments'
PR_URL = "http://example.com/projects/PRO/repos/myrepo/pull-requests/20"


class TestBitbucketServerPRCommentPush(
    TestReactorMixin, unittest.TestCase, ReporterTestMixin, LoggingMixin
):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)
        yield self.master.startService()
        self.addCleanup(self.master.stopService)

    @defer.inlineCallbacks
    def setupReporter(self, verbose=True, generator_class=BuildStatusGenerator, **kwargs):
        self._http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, 'serv', auth=('username', 'passwd'), debug=None, verify=None
        )

        formatter = Mock(spec=MessageFormatter)
        formatter.format_message_for_build.return_value = {
            "body": UNICODE_BODY,
            "type": "text",
            "subject": "subject",
            "extra_info": None,
        }
        formatter.want_properties = True
        formatter.want_steps = False
        formatter.want_logs = False
        formatter.want_logs_content = False

        generator = generator_class(message_formatter=formatter)

        self.cp = BitbucketServerPRCommentPush(
            "serv",
            Interpolate("username"),
            Interpolate("passwd"),
            verbose=verbose,
            generators=[generator],
            **kwargs,
        )
        yield self.cp.setServiceParent(self.master)

    @defer.inlineCallbacks
    def setupBuildResults(self, buildResults, set_pr=True):
        yield super().insert_test_data([buildResults], buildResults)
        build = yield self.master.data.get(('builds', 20))
        if set_pr:
            yield self.master.db.builds.setBuildProperty(20, "pullrequesturl", PR_URL, "test")
        return build

    @defer.inlineCallbacks
    def test_reporter_basic(self):
        yield self.setupReporter()
        build = yield self.setupBuildResults(SUCCESS)
        self._http.expect("post", EXPECTED_API, json={"text": UNICODE_BODY}, code=HTTP_CREATED)
        build["complete"] = True
        self.setUpLogging()
        yield self.cp._got_event(('builds', 20, 'finished'), build)
        self.assertLogged(f'Comment sent to {PR_URL}')

    @defer.inlineCallbacks
    def test_reporter_basic_without_logging(self):
        yield self.setupReporter(verbose=False)
        build = yield self.setupBuildResults(SUCCESS)
        self._http.expect("post", EXPECTED_API, json={"text": UNICODE_BODY}, code=HTTP_CREATED)
        build["complete"] = True
        self.setUpLogging()
        yield self.cp._got_event(('builds', 20, 'finished'), build)

        self.assertNotLogged(f'Comment sent to {PR_URL}')

    @defer.inlineCallbacks
    def test_reporter_without_pullrequest(self):
        yield self.setupReporter()
        build = yield self.setupBuildResults(SUCCESS, set_pr=False)
        build["complete"] = True
        # we don't expect any request
        yield self.cp._got_event(('builds', 20, 'finished'), build)

    @defer.inlineCallbacks
    def test_reporter_with_buildset(self):
        yield self.setupReporter(generator_class=BuildSetStatusGenerator)
        yield self.setupBuildResults(SUCCESS)
        buildset = yield self.get_inserted_buildset()
        self._http.expect("post", EXPECTED_API, json={"text": UNICODE_BODY}, code=HTTP_CREATED)
        yield self.cp._got_event(("buildsets", 98, "complete"), buildset)

    @defer.inlineCallbacks
    def test_reporter_logs_error_code_and_content_on_invalid_return_code(self):
        yield self.setupReporter()
        build = yield self.setupBuildResults(SUCCESS)

        http_error_code = 500
        error_body = {"errors": [{"message": "A dataXXXbase error has occurred."}]}

        self._http.expect(
            "post",
            EXPECTED_API,
            json={"text": UNICODE_BODY},
            code=http_error_code,
            content_json=error_body,
        )
        self.setUpLogging()
        build['complete'] = True
        yield self.cp._got_event(('builds', 20, 'finished'), build)

        self.assertLogged(f"^{http_error_code}: Unable to send a comment: ")
        self.assertLogged("A dataXXXbase error has occurred")

    @defer.inlineCallbacks
    def test_reporter_logs_error_code_without_content_on_invalid_return_code(self):
        yield self.setupReporter()
        build = yield self.setupBuildResults(SUCCESS)
        http_error_code = 503
        self._http.expect("post", EXPECTED_API, json={"text": UNICODE_BODY}, code=http_error_code)
        self.setUpLogging()
        build['complete'] = True
        yield self.cp._got_event(('builds', 20, 'finished'), build)
        self.assertLogged(f"^{http_error_code}: Unable to send a comment: ")

    @defer.inlineCallbacks
    def test_reporter_does_not_log_return_code_on_valid_return_code(self):
        yield self.setupReporter()
        build = yield self.setupBuildResults(SUCCESS)
        http_code = 201
        self._http.expect("post", EXPECTED_API, json={"text": UNICODE_BODY}, code=http_code)
        self.setUpLogging()
        build['complete'] = True
        yield self.cp._got_event(('builds', 20, 'finished'), build)
        self.assertNotLogged(f"^{http_code}:")



#!/usr/bin/env python
#
# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Standard setup script.
"""

from setuptools import setup  # isort:skip


import inspect
import os
import sys

from setuptools import Command
from setuptools.command.sdist import sdist

from buildbot import version

BUILDING_WHEEL = bool("bdist_wheel" in sys.argv)


class install_data_twisted(Command):
    """make sure VERSION file is installed in package."""

    def initialize_options(self):
        self.install_dir = None

    def finalize_options(self):
        self.set_undefined_options(
            'install',
            ('install_lib', 'install_dir'),
        )

    def run(self):
        # ensure there's a buildbot/VERSION file
        fn = os.path.join(self.install_dir, 'buildbot', 'VERSION')
        with open(fn, 'w') as f:
            f.write(version)


class our_sdist(sdist):
    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)

        # ensure there's a buildbot/VERSION file
        fn = os.path.join(base_dir, 'buildbot', 'VERSION')
        with open(fn, 'w') as f:
            f.write(version)

        # ensure that NEWS has a copy of the latest release notes, with the
        # proper version substituted
        src_fn = os.path.join('docs', 'relnotes/index.rst')
        with open(src_fn) as f:
            src = f.read()
        src = src.replace('|version|', version)
        dst_fn = os.path.join(base_dir, 'NEWS')
        with open(dst_fn, 'w') as f:
            f.write(src)


def define_plugin_entry(name, module_name):
    """
    helper to produce lines suitable for setup.py's entry_points
    """
    if isinstance(name, tuple):
        entry, name = name
    else:
        entry = name
    return f'{entry} = {module_name}:{name}'


def concat_dicts(*dicts):
    result = {}
    for d in dicts:
        result.update(d)
    return result


def define_plugin_entries(groups):
    """
    helper to all groups for plugins
    """
    result = {}

    for group, modules in groups:
        tempo = []
        for module_name, names in modules:
            tempo.extend([define_plugin_entry(name, module_name) for name in names])
        result[group] = tempo

    return result


__file__ = inspect.getframeinfo(inspect.currentframe()).filename

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as long_d_f:
    long_description = long_d_f.read()

setup_args = {
    'name': "buildbot",
    'version': version,
    'description': "The Continuous Integration Framework",
    'long_description': long_description,
    'author': "Brian Warner",
    'author_email': "warner-buildbot@lothar.com",
    'maintainer': "Dustin J. Mitchell",
    'maintainer_email': "dustin@v.igoro.us",
    'url': "http://buildbot.net/",
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Environment :: No Input/Output (Daemon)',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    'packages': [
        "buildbot",
        "buildbot.configurators",
        "buildbot.worker",
        "buildbot.worker.protocols",
        "buildbot.worker.protocols.manager",
        "buildbot.changes",
        "buildbot.clients",
        "buildbot.config",
        "buildbot.data",
        "buildbot.db",
        "buildbot.db.compression",
        "buildbot.db.migrations",
        "buildbot.db.migrations.versions",
        "buildbot.db.types",
        "buildbot.machine",
        "buildbot.monkeypatches",
        "buildbot.mq",
        "buildbot.plugins",
        "buildbot.process",
        "buildbot.process.users",
        "buildbot.reporters",
        "buildbot.reporters.generators",
        "buildbot.schedulers",
        "buildbot.scripts",
        "buildbot.secrets",
        "buildbot.secrets.providers",
        "buildbot.spec",
        "buildbot.spec.types",
        "buildbot.statistics",
        "buildbot.statistics.storage_backends",
        "buildbot.steps",
        "buildbot.steps.package",
        "buildbot.steps.package.deb",
        "buildbot.steps.package.rpm",
        "buildbot.steps.source",
        "buildbot.util",
        "buildbot.wamp",
        "buildbot.www",
        "buildbot.www.hooks",
        "buildbot.www.authz",
        "buildbot.test",
        "buildbot.test.util",
        "buildbot.test.fake",
        "buildbot.test.fakedb",
        "buildbot.test.integration.pki",
        "buildbot.test.integration.pki.ca",
        "buildbot.test.unit.test_templates_dir",
        "buildbot.test.unit.test_templates_dir.plugin",
    ]
    + (
        []
        if BUILDING_WHEEL
        else [  # skip tests for wheels (save 50% of the archive)
            "buildbot.test.fuzz",
            "buildbot.test.integration",
            "buildbot.test.integration.interop",
            "buildbot.test.regressions",
            "buildbot.test.unit",
        ]
    ),
    # mention data_files, even if empty, so install_data is called and
    # VERSION gets copied
    'data_files': [("buildbot", [])],
    'package_data': {
        "": ["VERSION"],
        "buildbot.reporters.templates": ["*.txt"],
        "buildbot.db.migrations": [
            "alembic.ini",
            "README",
        ],
        "buildbot.db.migrations.versions": ["*.py"],
        "buildbot.scripts": [
            "sample.cfg",
            "buildbot_tac.tmpl",
        ],
        "buildbot.spec": ["*.raml"],
        "buildbot.spec.types": ["*.raml"],
        "buildbot.test.unit.test_templates_dir": ["*.html"],
        "buildbot.test.unit.test_templates_dir.plugin": ["*.*"],
        "buildbot.test.integration.pki": ["*.*"],
        "buildbot.test.integration.pki.ca": ["*.*"],
    },
    'cmdclass': {'install_data': install_data_twisted, 'sdist': our_sdist},
    'entry_points': concat_dicts(
        define_plugin_entries([
            (
                'buildbot.changes',
                [
                    (
                        'buildbot.changes.mail',
                        [
                            'MaildirSource',
                            'CVSMaildirSource',
                            'SVNCommitEmailMaildirSource',
                            'BzrLaunchpadEmailMaildirSource',
                        ],
                    ),
                    ('buildbot.changes.bitbucket', ['BitbucketPullrequestPoller']),
                    ('buildbot.changes.github', ['GitHubPullrequestPoller']),
                    (
                        'buildbot.changes.gerritchangesource',
                        ['GerritChangeSource', 'GerritEventLogPoller'],
                    ),
                    ('buildbot.changes.gitpoller', ['GitPoller']),
                    ('buildbot.changes.hgpoller', ['HgPoller']),
                    ('buildbot.changes.p4poller', ['P4Source']),
                    ('buildbot.changes.pb', ['PBChangeSource']),
                    ('buildbot.changes.svnpoller', ['SVNPoller']),
                ],
            ),
            (
                'buildbot.schedulers',
                [
                    ('buildbot.schedulers.basic', ['SingleBranchScheduler', 'AnyBranchScheduler']),
                    ('buildbot.schedulers.dependent', ['Dependent']),
                    ('buildbot.schedulers.triggerable', ['Triggerable']),
                    ('buildbot.schedulers.forcesched', ['ForceScheduler']),
                    ('buildbot.schedulers.timed', ['Periodic', 'Nightly', 'NightlyTriggerable']),
                    ('buildbot.schedulers.trysched', ['Try_Jobdir', 'Try_Userpass']),
                ],
            ),
            (
                'buildbot.secrets',
                [
                    ('buildbot.secrets.providers.file', ['SecretInAFile']),
                    ('buildbot.secrets.providers.passwordstore', ['SecretInPass']),
                    (
                        'buildbot.secrets.providers.vault_hvac',
                        [
                            'HashiCorpVaultKvSecretProvider',
                            'VaultAuthenticatorToken',
                            'VaultAuthenticatorApprole',
                        ],
                    ),
                ],
            ),
            (
                'buildbot.worker',
                [
                    ('buildbot.worker.base', ['Worker']),
                    ('buildbot.worker.ec2', ['EC2LatentWorker']),
                    ('buildbot.worker.libvirt', ['LibVirtWorker']),
                    ('buildbot.worker.openstack', ['OpenStackLatentWorker']),
                    ('buildbot.worker.docker', ['DockerLatentWorker']),
                    ('buildbot.worker.kubernetes', ['KubeLatentWorker']),
                    ('buildbot.worker.local', ['LocalWorker']),
                ],
            ),
            (
                'buildbot.machine',
                [
                    ('buildbot.machine.base', ['Machine']),
                ],
            ),
            (
                'buildbot.steps',
                [
                    ('buildbot.process.buildstep', ['BuildStep']),
                    ('buildbot.steps.cmake', ['CMake']),
                    (
                        'buildbot.steps.configurable',
                        ['BuildbotTestCiTrigger', 'BuildbotCiSetupSteps'],
                    ),
                    ('buildbot.steps.cppcheck', ['Cppcheck']),
                    ('buildbot.steps.gitdiffinfo', ['GitDiffInfo']),
                    (
                        'buildbot.steps.http',
                        ['HTTPStep', 'POST', 'GET', 'PUT', 'DELETE', 'HEAD', 'OPTIONS'],
                    ),
                    (
                        'buildbot.steps.master',
                        [
                            'MasterShellCommand',
                            'SetProperty',
                            'SetProperties',
                            'LogRenderable',
                            "Assert",
                        ],
                    ),
                    ('buildbot.steps.maxq', ['MaxQ']),
                    ('buildbot.steps.mswin', ['Robocopy']),
                    ('buildbot.steps.package.deb.lintian', ['DebLintian']),
                    (
                        'buildbot.steps.package.deb.pbuilder',
                        ['DebPbuilder', 'DebCowbuilder', 'UbuPbuilder', 'UbuCowbuilder'],
                    ),
                    ('buildbot.steps.package.rpm.mock', ['Mock', 'MockBuildSRPM', 'MockRebuild']),
                    ('buildbot.steps.package.rpm.rpmbuild', ['RpmBuild']),
                    ('buildbot.steps.package.rpm.rpmlint', ['RpmLint']),
                    ('buildbot.steps.python', ['BuildEPYDoc', 'PyFlakes', 'PyLint', 'Sphinx']),
                    ('buildbot.steps.python_twisted', ['HLint', 'Trial', 'RemovePYCs']),
                    (
                        'buildbot.steps.shell',
                        [
                            'ShellCommand',
                            'TreeSize',
                            'SetPropertyFromCommand',
                            'Configure',
                            'WarningCountingShellCommand',
                            'Compile',
                            'Test',
                            'PerlModuleTest',
                        ],
                    ),
                    ('buildbot.steps.shellsequence', ['ShellSequence']),
                    ('buildbot.steps.source.bzr', ['Bzr']),
                    ('buildbot.steps.source.cvs', ['CVS']),
                    ('buildbot.steps.source.darcs', ['Darcs']),
                    ('buildbot.steps.source.gerrit', ['Gerrit']),
                    ('buildbot.steps.source.git', ['Git', 'GitCommit', 'GitPush', 'GitTag']),
                    ('buildbot.steps.source.github', ['GitHub']),
                    ('buildbot.steps.source.gitlab', ['GitLab']),
                    ('buildbot.steps.source.mercurial', ['Mercurial']),
                    ('buildbot.steps.source.mtn', ['Monotone']),
                    ('buildbot.steps.source.p4', ['P4']),
                    ('buildbot.steps.source.repo', ['Repo']),
                    ('buildbot.steps.source.svn', ['SVN']),
                    ('buildbot.steps.subunit', ['SubunitShellCommand']),
                    (
                        'buildbot.steps.transfer',
                        [
                            'FileUpload',
                            'DirectoryUpload',
                            'MultipleFileUpload',
                            'FileDownload',
                            'StringDownload',
                            'JSONStringDownload',
                            'JSONPropertiesDownload',
                        ],
                    ),
                    ('buildbot.steps.trigger', ['Trigger']),
                    (
                        'buildbot.steps.vstudio',
                        [
                            'VC6',
                            'VC7',
                            'VS2003',
                            'VC8',
                            'VS2005',
                            'VCExpress9',
                            'VC9',
                            'VS2008',
                            'VC10',
                            'VS2010',
                            'VC11',
                            'VS2012',
                            'VC12',
                            'VS2013',
                            'VC14',
                            'VS2015',
                            'VC141',
                            'VS2017',
                            'VS2019',
                            'VS2022',
                            'MsBuild4',
                            'MsBuild',
                            'MsBuild12',
                            'MsBuild14',
                            'MsBuild141',
                            'MsBuild15',
                            'MsBuild16',
                            'MsBuild17',
                        ],
                    ),
                    (
                        'buildbot.steps.worker',
                        [
                            'SetPropertiesFromEnv',
                            'FileExists',
                            'CopyDirectory',
                            'RemoveDirectory',
                            'MakeDirectory',
                        ],
                    ),
                ],
            ),
            (
                'buildbot.reporters',
                [
                    (
                        'buildbot.reporters.generators.build',
                        ['BuildStatusGenerator', 'BuildStartEndStatusGenerator'],
                    ),
                    ('buildbot.reporters.generators.buildrequest', ['BuildRequestGenerator']),
                    (
                        "buildbot.reporters.generators.buildset",
                        [
                            "BuildSetCombinedStatusGenerator",
                            "BuildSetStatusGenerator",
                        ],
                    ),
                    ('buildbot.reporters.generators.worker', ['WorkerMissingGenerator']),
                    ('buildbot.reporters.mail', ['MailNotifier']),
                    ('buildbot.reporters.pushjet', ['PushjetNotifier']),
                    ('buildbot.reporters.pushover', ['PushoverNotifier']),
                    (
                        'buildbot.reporters.message',
                        [
                            'MessageFormatter',
                            'MessageFormatterEmpty',
                            'MessageFormatterFunction',
                            "MessageFormatterFunctionRaw",
                            'MessageFormatterMissingWorker',
                            'MessageFormatterRenderable',
                        ],
                    ),
                    ('buildbot.reporters.gerrit', ['GerritStatusPush']),
                    ('buildbot.reporters.gerrit_verify_status', ['GerritVerifyStatusPush']),
                    ('buildbot.reporters.http', ['HttpStatusPush']),
                    ('buildbot.reporters.github', ['GitHubStatusPush', 'GitHubCommentPush']),
                    ('buildbot.reporters.gitlab', ['GitLabStatusPush']),
                    (
                        'buildbot.reporters.bitbucketserver',
                        [
                            'BitbucketServerStatusPush',
                            'BitbucketServerCoreAPIStatusPush',
                            'BitbucketServerPRCommentPush',
                        ],
                    ),
                    ('buildbot.reporters.bitbucket', ['BitbucketStatusPush']),
                    ('buildbot.reporters.irc', ['IRC']),
                    ('buildbot.reporters.telegram', ['TelegramBot']),
                    ('buildbot.reporters.zulip', ['ZulipStatusPush']),
                ],
            ),
            (
                'buildbot.util',
                [
                    # Connection seems to be a way too generic name, though
                    ('buildbot.worker.libvirt', ['Connection']),
                    ('buildbot.changes.filter', ['ChangeFilter']),
                    ('buildbot.changes.gerritchangesource', ['GerritChangeFilter']),
                    (
                        'buildbot.changes.svnpoller',
                        [
                            ('svn.split_file_projects_branches', 'split_file_projects_branches'),
                            ('svn.split_file_branches', 'split_file_branches'),
                            ('svn.split_file_alwaystrunk', 'split_file_alwaystrunk'),
                        ],
                    ),
                    ('buildbot.configurators.janitor', ['JanitorConfigurator']),
                    ('buildbot.config.builder', ['BuilderConfig']),
                    (
                        'buildbot.locks',
                        [
                            'MasterLock',
                            'WorkerLock',
                        ],
                    ),
                    (
                        'buildbot.manhole',
                        ['AuthorizedKeysManhole', 'PasswordManhole', 'TelnetManhole'],
                    ),
                    (
                        'buildbot.process.builder',
                        [
                            'enforceChosenWorker',
                        ],
                    ),
                    (
                        'buildbot.process.codebase',
                        [
                            'Codebase',
                        ],
                    ),
                    (
                        'buildbot.process.factory',
                        [
                            'BuildFactory',
                            'GNUAutoconf',
                            'CPAN',
                            'Distutils',
                            'Trial',
                            'BasicBuildFactory',
                            'QuickBuildFactory',
                            'BasicSVN',
                        ],
                    ),
                    ('buildbot.process.logobserver', ['LogLineObserver']),
                    ('buildbot.process.project', ['Project']),
                    (
                        'buildbot.process.properties',
                        [
                            'FlattenList',
                            'Interpolate',
                            'Property',
                            'Transform',
                            'WithProperties',
                            'renderer',
                            'Secret',
                        ],
                    ),
                    ('buildbot.process.users.manual', ['CommandlineUserManager']),
                    ('buildbot.revlinks', ['RevlinkMatch']),
                    ('buildbot.reporters.utils', ['URLForBuild']),
                    ('buildbot.schedulers.canceller', ['OldBuildCanceller']),
                    ('buildbot.schedulers.canceller_buildset', ['FailingBuildsetCanceller']),
                    (
                        'buildbot.schedulers.forcesched',
                        [
                            'AnyPropertyParameter',
                            'BooleanParameter',
                            'ChoiceStringParameter',
                            'CodebaseParameter',
                            'FileParameter',
                            'FixedParameter',
                            'InheritBuildParameter',
                            'IntParameter',
                            'NestedParameter',
                            'ParameterGroup',
                            'PatchParameter',
                            'StringParameter',
                            'TextParameter',
                            'UserNameParameter',
                            'WorkerChoiceParameter',
                        ],
                    ),
                    (
                        'buildbot.process.results',
                        [
                            'Results',
                            'SUCCESS',
                            'WARNINGS',
                            'FAILURE',
                            'SKIPPED',
                            'EXCEPTION',
                            'RETRY',
                            'CANCELLED',
                        ],
                    ),
                    (
                        'buildbot.steps.source.repo',
                        [
                            ('repo.DownloadsFromChangeSource', 'RepoDownloadsFromChangeSource'),
                            ('repo.DownloadsFromProperties', 'RepoDownloadsFromProperties'),
                        ],
                    ),
                    ('buildbot.steps.shellsequence', ['ShellArg']),
                    (
                        'buildbot.util.git_credential',
                        ['GitCredentialInputRenderer', 'GitCredentialOptions'],
                    ),
                    (
                        'buildbot.util.kubeclientservice',
                        [
                            'KubeHardcodedConfig',
                            'KubeCtlProxyConfigLoader',
                            'KubeInClusterConfigLoader',
                        ],
                    ),
                    ('buildbot.util.ssfilter', ['SourceStampFilter']),
                    ('buildbot.www.avatar', ['AvatarGravatar', 'AvatarGitHub']),
                    (
                        'buildbot.www.auth',
                        ['UserPasswordAuth', 'HTPasswdAuth', 'RemoteUserAuth', 'CustomAuth'],
                    ),
                    ('buildbot.www.ldapuserinfo', ['LdapUserInfo']),
                    (
                        'buildbot.www.oauth2',
                        ['GoogleAuth', 'GitHubAuth', 'GitLabAuth', 'BitbucketAuth'],
                    ),
                    ('buildbot.db.dbconfig', ['DbConfig']),
                    ('buildbot.www.authz', ['Authz', 'fnmatchStrMatcher', 'reStrMatcher']),
                    (
                        'buildbot.www.authz.roles',
                        [
                            'RolesFromEmails',
                            'RolesFromGroups',
                            'RolesFromOwner',
                            'RolesFromUsername',
                            'RolesFromDomain',
                        ],
                    ),
                    (
                        'buildbot.www.authz.endpointmatchers',
                        [
                            'AnyEndpointMatcher',
                            'StopBuildEndpointMatcher',
                            'ForceBuildEndpointMatcher',
                            'RebuildBuildEndpointMatcher',
                            'AnyControlEndpointMatcher',
                            'EnableSchedulerEndpointMatcher',
                        ],
                    ),
                ],
            ),
            (
                'buildbot.webhooks',
                [
                    ('buildbot.www.hooks.base', ['base']),
                    ('buildbot.www.hooks.bitbucket', ['bitbucket']),
                    ('buildbot.www.hooks.github', ['github']),
                    ('buildbot.www.hooks.gitlab', ['gitlab']),
                    ('buildbot.www.hooks.gitorious', ['gitorious']),
                    ('buildbot.www.hooks.poller', ['poller']),
                    ('buildbot.www.hooks.bitbucketcloud', ['bitbucketcloud']),
                    ('buildbot.www.hooks.bitbucketserver', ['bitbucketserver']),
                ],
            ),
        ]),
        {
            'console_scripts': [
                'buildbot=buildbot.scripts.runner:run',
                # this will also be shipped on non windows :-(
                'buildbot_windows_service=buildbot.scripts.windows_service:HandleCommandLine',
            ]
        },
    ),
}

# set zip_safe to false to force Windows installs to always unpack eggs
# into directories, which seems to work better --
# see http://buildbot.net/trac/ticket/907
if sys.platform == "win32":
    setup_args['zip_safe'] = False

py_38 = sys.version_info[0] > 3 or (sys.version_info[0] == 3 and sys.version_info[1] >= 8)
if not py_38:
    raise RuntimeError("Buildbot master requires at least Python-3.8")

twisted_ver = ">= 22.1.0"

bundle_version = version.split("-")[0]

# dependencies
setup_args['install_requires'] = [
    'setuptools >= 8.0',
    'Twisted ' + twisted_ver,
    'treq >= 20.9',
    'Jinja2 >= 2.1',
    'msgpack >= 0.6.0',
    "croniter >= 1.3.0",
    'importlib-resources >= 5; python_version < "3.9"',
    # required for tests, but Twisted requires this anyway
    'zope.interface >= 4.1.1',
    'sqlalchemy >= 1.4.0',
    'alembic >= 1.6.0',
    'python-dateutil>=1.5',
    "txaio >= 2.2.2",
    "autobahn >= 0.16.0",
    'packaging',
    'PyJWT',
    'pyyaml',
    'unidiff >= 0.7.5',
]

# buildbot_windows_service needs pywin32
if sys.platform == "win32":
    setup_args['install_requires'].append('pywin32')

# Unit test dependencies.
test_deps = [
    # http client libraries
    'treq',
    'txrequests',
    # pypugjs required for custom templates tests
    'pypugjs',
    # boto3 and moto required for running EC2 tests
    'boto3',
    'moto',
    "Markdown>=3.0.0",
    'parameterized',
]
if sys.platform != 'win32':
    test_deps += [
        # LZ4 fails to build on Windows:
        # https://github.com/steeve/python-lz4/issues/27
        # lz4 required for log compression tests.
        'lz4',
    ]

setup_args['tests_require'] = test_deps

setup_args['extras_require'] = {
    'test': ["setuptools_trial", "ruff", *test_deps],
    'bundle': [
        f"buildbot-www=={bundle_version}",
        f"buildbot-worker=={bundle_version}",
        f"buildbot-waterfall-view=={bundle_version}",
        f"buildbot-console-view=={bundle_version}",
        f"buildbot-grid-view=={bundle_version}",
    ],
    'tls': [
        'Twisted[tls] ' + twisted_ver,
        # There are bugs with extras inside extras:
        # <https://github.com/pypa/pip/issues/3516>
        # so we explicitly include Twisted[tls] dependencies.
        'pyopenssl >= 16.0.0',
        'service_identity',
        'idna >= 0.6',
    ],
    'docs': [
        'docutils>=0.16.0',
        'sphinx>=3.2.0',
        'sphinx-rtd-theme>=0.5',
        'sphinxcontrib-spelling',
        'sphinxcontrib-websupport',
        'pyenchant',
        'sphinx-jinja',
        'towncrier',
    ],
    'brotli': [
        'Brotli>=1.1.0',
    ],
    'zstd': [
        'zstandard>=0.23.0',
    ],
    'configurable': [
        'evalidate >= 2.0.0',
    ],
}

if '--help-commands' in sys.argv or 'trial' in sys.argv or 'test' in sys.argv:
    setup_args['setup_requires'] = [
        'setuptools_trial',
    ]

if os.getenv('NO_INSTALL_REQS'):
    setup_args['install_requires'] = None
    setup_args['extras_require'] = None

if __name__ == '__main__':
    setup(**setup_args)

# Local Variables:
# fill-column: 71
# End:



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class TestStateConnectorComponent(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_getObjectId_new(self):
        objectid = yield self.db.state.getObjectId('someobj', 'someclass')

        yield self.assertNotEqual(objectid, None)

        def thd(conn):
            q = self.db.model.objects.select()
            rows = conn.execute(q).fetchall()
            self.assertEqual(
                [(r.id, r.name, r.class_name) for r in rows], [(objectid, 'someobj', 'someclass')]
            )

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_getObjectId_existing(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=19, name='someobj', class_name='someclass')
        ])
        objectid = yield self.db.state.getObjectId('someobj', 'someclass')

        self.assertEqual(objectid, 19)

    @defer.inlineCallbacks
    def test_getObjectId_conflict(self):
        # set up to insert a row between looking for an existing object
        # and adding a new one, triggering the fallback to re-running
        # the select.
        def hook(conn):
            conn.execute(
                self.db.model.objects.insert().values(id=27, name='someobj', class_name='someclass')
            )
            conn.commit()

        self.db.state._test_timing_hook = hook

        objectid = yield self.db.state.getObjectId('someobj', 'someclass')

        self.assertEqual(objectid, 27)

    @defer.inlineCallbacks
    def test_getObjectId_new_big_name(self):
        objectid = yield self.db.state.getObjectId('someobj' * 150, 'someclass')
        expn = 'someobj' * 9 + 's132bf9b89b0cdbc040d1ebc69e0dbee85dff720a'

        self.assertNotEqual(objectid, None)

        def thd(conn):
            q = self.db.model.objects.select()
            rows = conn.execute(q).fetchall()
            self.assertEqual(
                [(r.id, r.name, r.class_name) for r in rows], [(objectid, expn, 'someclass')]
            )

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_getState_missing(self):
        with self.assertRaises(KeyError):
            yield self.db.state.getState(10, 'nosuch')
        self.flushLoggedErrors(KeyError)

    @defer.inlineCallbacks
    def test_getState_missing_default(self):
        val = yield self.db.state.getState(10, 'nosuch', 'abc')

        self.assertEqual(val, 'abc')

    @defer.inlineCallbacks
    def test_getState_missing_default_None(self):
        val = yield self.db.state.getState(10, 'nosuch', None)

        self.assertEqual(val, None)

    @defer.inlineCallbacks
    def test_getState_present(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='x', class_name='y'),
            fakedb.ObjectState(objectid=10, name='x', value_json='[1,2]'),
        ])
        val = yield self.db.state.getState(10, 'x')

        self.assertEqual(val, [1, 2])

    @defer.inlineCallbacks
    def test_getState_badjson(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='x', class_name='y'),
            fakedb.ObjectState(objectid=10, name='x', value_json='ff[1'),
        ])
        with self.assertRaises(TypeError):
            yield self.db.state.getState(10, 'x')
        self.flushLoggedErrors(TypeError)

    @defer.inlineCallbacks
    def test_setState(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
        ])
        yield self.db.state.setState(10, 'x', [1, 2])

        def thd(conn):
            q = self.db.model.object_state.select()
            rows = conn.execute(q).fetchall()
            self.assertEqual(
                [(r.objectid, r.name, r.value_json) for r in rows], [(10, 'x', '[1, 2]')]
            )

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_setState_badjson(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='x', class_name='y'),
        ])
        with self.assertRaises(TypeError):
            yield self.db.state.setState(10, 'x', self)  # self is not JSON-able..
        self.flushLoggedErrors(TypeError)

    @defer.inlineCallbacks
    def test_setState_existing(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
            fakedb.ObjectState(objectid=10, name='x', value_json='99'),
        ])
        yield self.db.state.setState(10, 'x', [1, 2])

        def thd(conn):
            q = self.db.model.object_state.select()
            rows = conn.execute(q).fetchall()
            self.assertEqual(
                [(r.objectid, r.name, r.value_json) for r in rows], [(10, 'x', '[1, 2]')]
            )

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_setState_conflict(self):
        def hook(conn):
            conn.execute(
                self.db.model.object_state.insert().values(objectid=10, name='x', value_json='22')
            )
            conn.commit()

        self.db.state._test_timing_hook = hook

        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
        ])
        yield self.db.state.setState(10, 'x', [1, 2])

        def thd(conn):
            q = self.db.model.object_state.select()
            rows = conn.execute(q).fetchall()
            self.assertEqual([(r.objectid, r.name, r.value_json) for r in rows], [(10, 'x', '22')])

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_atomicCreateState(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
        ])
        res = yield self.db.state.atomicCreateState(10, 'x', lambda: [1, 2])
        self.assertEqual(res, [1, 2])
        res = yield self.db.state.getState(10, 'x')
        self.assertEqual(res, [1, 2])

    @defer.inlineCallbacks
    def test_atomicCreateState_conflict(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
        ])

        def hook(conn):
            conn.execute(
                self.db.model.object_state.insert().values(objectid=10, name='x', value_json='22')
            )
            conn.commit()

        self.db.state._test_timing_hook = hook

        res = yield self.db.state.atomicCreateState(10, 'x', lambda: [1, 2])
        self.assertEqual(res, 22)
        res = yield self.db.state.getState(10, 'x')
        self.assertEqual(res, 22)

    @defer.inlineCallbacks
    def test_atomicCreateState_nojsonable(self):
        yield self.db.insert_test_data([
            fakedb.Object(id=10, name='-', class_name='-'),
        ])

        with self.assertRaises(TypeError):
            yield self.db.state.atomicCreateState(10, 'x', object)
        self.flushLoggedErrors(TypeError)
